import json
import os

from ai.chronon.scheduler.interfaces.flow import Flow
from ai.chronon.scheduler.interfaces.node import Node
from ai.chronon.scheduler.interfaces.orchestrator import WorkflowOrchestrator
from ai.chronon.utils import (
    convert_json_to_obj,
    dict_to_bash_commands,
    dict_to_exports,
    get_config_path,
    get_join_output_table_name,
    join_part_name,
    sanitize,
)

TASK_PREFIX = "compute_join"
DEFAULT_SPARK_SETTINGS = {
    "default": {
        "spark_version": "3.1.1",
        "executor_memory": "4G",
        "driver_memory": "4G",
        "executor_cores": 2,
    }
}


class JoinBackfill:
    def __init__(
        self,
        start_date: str,
        end_date: str,
        config_path: str,
        extra_args: dict = None,
        settings: dict = None,
    ):
        self.dag_id = "_".join(
            map(
                sanitize, ["chronon_joins_backfill", os.path.basename(config_path).split("/")[-1], start_date, end_date]
            )
        )
        self.start_date = start_date
        self.end_date = end_date
        self.config_path = config_path
        self.extra_args = extra_args or {}
        self.settings = settings or DEFAULT_SPARK_SETTINGS
        with open(self.config_path, "r") as file:
            config = file.read()
        self.join = convert_json_to_obj(json.loads(config))

    def build_flow(self) -> Flow:
        flow = Flow(self.join.metaData.name)
        self.add_join_to_flow(flow, self.join)
        return flow

    def add_join_to_flow(self, flow: Flow, join: dict) -> Node:
        """
        Parse a Join object and add it to the Flow.
        Each join part is a node and will run in parallel. Besides, left table and final join are also nodes.
        We will add Join recursively when we encounter a join source.
        The final join node will be returned to be used as a dependency.
        """
        join_name = sanitize(join.metaData.name)
        final_node = Node(
            f"{TASK_PREFIX}__{sanitize(get_join_output_table_name(join, full_name=True))}",
            self.run_final_join(join.metaData.name),
        )
        left_node = Node(f"{TASK_PREFIX}__{join_name}__left_table", self.run_left_table(join.metaData.name))
        flow.add_node(final_node)
        flow.add_node(left_node)
        for join_part in join.joinParts:
            for src in join_part.groupBy.sources:
                if "joinSource" in src:
                    dep_node = self.add_join_to_flow(flow, src.joinSource.join)
                    left_node.add_dependency(dep_node)
            jp_full_name = join_part_name(join_part)
            jp_node = Node(
                f"{TASK_PREFIX}__{join_name}__{jp_full_name}", self.run_join_part(join.metaData.name, jp_full_name)
            )
            flow.add_node(jp_node)
            jp_node.add_dependency(left_node)
            final_node.add_dependency(jp_node)
        return final_node

    def export_template(self, settings: dict):
        return f"{dict_to_exports(settings)}"

    def command_template(self, config_path: str, extra_args: dict):
        extra_args.update(self.extra_args)
        if self.start_date:
            extra_args.update({"start_ds": self.start_date})
        return f"""run.py --conf={config_path} --ds={self.end_date} \
{dict_to_bash_commands(extra_args)}"""

    def run_join_part(self, join_name: dict, join_part: str):
        args = {
            "mode": "backfill",
            "selected_join_parts": join_part,
            "use_cached_left": None,
        }
        settings = self.settings.get(join_part, self.settings["default"])
        return (
            self.export_template(settings)
            + " && "
            + self.command_template(config_path=get_config_path(join_name), extra_args=args)
        )

    def run_left_table(self, join_name: str):
        settings = self.settings.get("left_table", self.settings["default"])
        return (
            self.export_template(settings)
            + " && "
            + self.command_template(config_path=get_config_path(join_name), extra_args={"mode": "backfill-left"})
        )

    def run_final_join(self, join_name: str):
        settings = self.settings.get("final_join", self.settings["default"])
        return (
            self.export_template(settings)
            + " && "
            + self.command_template(config_path=get_config_path(join_name), extra_args={"mode": "backfill-final"})
        )

    def run(self, orchestrator: WorkflowOrchestrator):
        orchestrator.setup()
        orchestrator.build_dag_from_flow(self.build_flow())
        orchestrator.trigger_run()




#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

JOIN_FOLDER_NAME = 'joins'
GROUP_BY_FOLDER_NAME = 'group_bys'
STAGING_QUERY_FOLDER_NAME = 'staging_queries'
# TODO - make team part of thrift API?
TEAMS_FILE_PATH = 'teams.json'



#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import inspect
import json
import logging
from typing import Callable, Dict, List, Optional, Tuple, Union

import ai.chronon.api.ttypes as ttypes
import ai.chronon.utils as utils

OperationType = int  # type(zthrift.Operation.FIRST)

#  The GroupBy's default online/production status is None and it will inherit
# online/production status from the Joins it is included.
# If it is included in multiple joins, it is considered online/production
# if any of the joins are online/production. Otherwise it is not online/production
# unless it is explicitly marked as online/production on the GroupBy itself.
DEFAULT_ONLINE = None
DEFAULT_PRODUCTION = None
LOGGER = logging.getLogger()


def collector(
    op: ttypes.Operation,
) -> Callable[[ttypes.Operation], Tuple[ttypes.Operation, Dict[str, str]]]:
    return lambda k: (op, {"k": str(k)})


def generic_collector(op: ttypes.Operation, required, **kwargs):
    def _collector(*args, **other_args):
        arguments = kwargs.copy() if kwargs else {}
        for idx, arg in enumerate(required):
            arguments[arg] = args[idx]
        arguments.update(other_args)
        return (op, {k: str(v) for k, v in arguments.items()})

    return _collector


# To simplify imports
class Accuracy(ttypes.Accuracy):
    pass


class Operation:
    MIN = ttypes.Operation.MIN
    MAX = ttypes.Operation.MAX
    FIRST = ttypes.Operation.FIRST
    LAST = ttypes.Operation.LAST
    APPROX_UNIQUE_COUNT = ttypes.Operation.APPROX_UNIQUE_COUNT
    # refer to the chart here to tune your sketch size with lgK
    # default is 8
    # https://github.com/apache/incubator-datasketches-java/blob/master/src/main/java/org/apache/datasketches/cpc/CpcSketch.java#L180
    APPROX_UNIQUE_COUNT_LGK = collector(ttypes.Operation.APPROX_UNIQUE_COUNT)
    UNIQUE_COUNT = ttypes.Operation.UNIQUE_COUNT
    COUNT = ttypes.Operation.COUNT
    SUM = ttypes.Operation.SUM
    AVERAGE = ttypes.Operation.AVERAGE
    VARIANCE = ttypes.Operation.VARIANCE
    SKEW = ttypes.Operation.SKEW
    KURTOSIS = ttypes.Operation.KURTOSIS
    HISTOGRAM = ttypes.Operation.HISTOGRAM
    # k truncates the map to top_k most frequent items, 0 turns off truncation
    HISTOGRAM_K = collector(ttypes.Operation.HISTOGRAM)
    # k truncates the map to top_k most frequent items, k is required and results are bounded
    APPROX_HISTOGRAM_K = collector(ttypes.Operation.APPROX_HISTOGRAM_K)
    FIRST_K = collector(ttypes.Operation.FIRST_K)
    LAST_K = collector(ttypes.Operation.LAST_K)
    TOP_K = collector(ttypes.Operation.TOP_K)
    BOTTOM_K = collector(ttypes.Operation.BOTTOM_K)
    APPROX_PERCENTILE = generic_collector(ttypes.Operation.APPROX_PERCENTILE, ["percentiles"], k=128)


def Aggregations(**agg_dict):
    assert all(isinstance(agg, ttypes.Aggregation) for agg in agg_dict.values())
    for key, agg in agg_dict.items():
        if not agg.inputColumn:
            agg.inputColumn = key
    return agg_dict.values()


def DefaultAggregation(keys, sources, operation=Operation.LAST, tags=None):
    aggregate_columns = []
    for source in sources:
        query = utils.get_query(source)
        columns = utils.get_columns(source)
        non_aggregate_columns = keys + [
            "ts",
            "is_before",
            "mutation_ts",
            "ds",
            query.timeColumn,
        ]
        aggregate_columns += [column for column in columns if column not in non_aggregate_columns]
    return [Aggregation(operation=operation, input_column=column, tags=tags) for column in aggregate_columns]


class TimeUnit:
    HOURS = ttypes.TimeUnit.HOURS
    DAYS = ttypes.TimeUnit.DAYS


def window_to_str_pretty(window: ttypes.Window):
    unit = ttypes.TimeUnit._VALUES_TO_NAMES[window.timeUnit].lower()
    return f"{window.length} {unit}"


def op_to_str(operation: OperationType):
    return ttypes.Operation._VALUES_TO_NAMES[operation].lower()


# See docs/Aggregations.md
def Aggregation(
    input_column: str = None,
    operation: Union[ttypes.Operation, Tuple[ttypes.Operation, Dict[str, str]]] = None,
    windows: List[ttypes.Window] = None,
    buckets: List[str] = None,
    tags: Dict[str, str] = None,
) -> ttypes.Aggregation:
    """
    :param input_column:
        Column on which the aggregation needs to be performed.
        This should be one of the input columns specified on the keys of the `select` in the `Query`'s `Source`
    :type input_column: str
    :param operation:
        Operation to use to aggregate the input columns. For example, MAX, MIN, COUNT
        Some operations have arguments, like last_k, approx_percentiles etc.,
        Defaults to "LAST".
    :type operation: ttypes.Operation
    :param windows:
        Length to window to calculate the aggregates on.
        Minimum window size is 1hr. Maximum can be arbitrary. When not defined, the computation is un-windowed.
    :type windows: List[ttypes.Window]
    :param buckets:
        Besides the GroupBy.keys, this is another level of keys for use under this aggregation.
        Using this would create an output as a map of string to aggregate.
    :type buckets: List[str]
    :return: An aggregate defined with the specified operation.
    """
    # Default to last
    operation = operation if operation is not None else Operation.LAST
    arg_map = {}
    if isinstance(operation, tuple):
        operation, arg_map = operation[0], operation[1]
    agg = ttypes.Aggregation(input_column, operation, arg_map, windows, buckets)
    agg.tags = tags
    return agg


def Window(length: int, timeUnit: ttypes.TimeUnit) -> ttypes.Window:
    return ttypes.Window(length, timeUnit)


def Derivation(name: str, expression: str) -> ttypes.Derivation:
    """
    Derivation allows arbitrary SQL select clauses to be computed using columns from the output of group by backfill
    output schema. It is supported for offline computations for now.

    If both name and expression are set to "*", then every raw column will be included along with the derived columns.

    :param name: output column name of the SQL expression
    :param expression: any valid Spark SQL select clause based on joinPart or externalPart columns
    :return: a Derivation object representing a single derived column or a wildcard ("*") selection.
    """
    return ttypes.Derivation(name=name, expression=expression)


def contains_windowed_aggregation(aggregations: Optional[List[ttypes.Aggregation]]):
    if not aggregations:
        return False
    for agg in aggregations:
        if agg.windows:
            return True
    return False


def validate_group_by(group_by: ttypes.GroupBy):
    sources = group_by.sources
    keys = group_by.keyColumns
    aggregations = group_by.aggregations
    # check ts is not included in query.select
    first_source_columns = set(utils.get_columns(sources[0]))
    # TODO undo this check after ml_models CI passes
    assert "ts" not in first_source_columns, (
        "'ts' is a reserved key word for Chronon," " please specify the expression in timeColumn"
    )
    for src in sources:
        query = utils.get_query(src)
        if src.events:
            assert query.mutationTimeColumn is None, (
                "ingestionTimeColumn should not be specified for "
                "event source as it should be the same with timeColumn"
            )
            assert query.reversalColumn is None, (
                "reversalColumn should not be specified for event source " "as it won't have mutations"
            )
            if group_by.accuracy != Accuracy.SNAPSHOT:
                assert query.timeColumn is not None, (
                    "please specify query.timeColumn for non-snapshot accurate " "group by with event source"
                )
        else:
            if contains_windowed_aggregation(aggregations):
                assert query.timeColumn, "Please specify timeColumn for entity source with windowed aggregations"

    column_set = None
    # all sources should select the same columns
    for i, source in enumerate(sources[1:]):
        column_set = set(utils.get_columns(source))
        column_diff = column_set ^ first_source_columns
        assert not column_diff, f"""
Mismatched columns among sources [1, {i+2}], Difference: {column_diff}
"""

    # all keys should be present in the selected columns
    unselected_keys = set(keys) - first_source_columns
    assert not unselected_keys, f"""
Keys {unselected_keys}, are unselected in source
"""

    # Aggregations=None is only valid if group_by is Entities
    if aggregations is None:
        is_events = any([s.events for s in sources])
        has_mutations = (
            any(
                [
                    (s.entities.mutationTable is not None or s.entities.mutationTopic is not None)
                    for s in sources
                    if s.entities is not None
                ]
            )
            if not is_events
            else False
        )
        assert not (
            is_events or has_mutations
        ), "You can only set aggregations=None in an EntitySource without mutations"
    else:
        columns = set([c for src in sources for c in utils.get_columns(src)])
        for agg in aggregations:
            assert agg.inputColumn, (
                f"input_column is required for all operations, found: input_column = {agg.inputColumn} "
                f"and operation {op_to_str(agg.operation)}"
            )
            assert (agg.inputColumn in columns) or (agg.inputColumn == "ts"), (
                f"input_column: for aggregation is not part of the query. Available columns: {column_set} "
                f"input_column: {agg.inputColumn}"
            )
            if agg.operation == ttypes.Operation.APPROX_PERCENTILE:
                if agg.argMap is not None and agg.argMap.get("percentiles") is not None:
                    try:
                        percentile_array = json.loads(agg.argMap["percentiles"])
                        assert isinstance(percentile_array, list)
                        assert all([float(p) >= 0 and float(p) <= 1 for p in percentile_array])
                    except Exception as e:
                        LOGGER.exception(e)
                        raise ValueError(
                            "[Percentiles] Unable to decode percentiles value, expected json array with values between"
                            f" 0 and 1 inclusive (ex: [0.6, 0.1]), received: {agg.argMap['percentiles']}"
                        )
                else:
                    raise ValueError(
                        f"[Percentiles] Unsupported arguments for {op_to_str(agg.operation)}, "
                        "example required: {'k': '128', 'percentiles': '[0.4,0.5,0.95]'},"
                        f" received: {agg.argMap}\n"
                    )
            if agg.windows:
                assert not (
                    # Snapshot accuracy.
                    ((group_by.accuracy and group_by.accuracy == Accuracy.SNAPSHOT) or group_by.backfillStartDate)
                    and
                    # Hourly aggregation.
                    any([window.timeUnit == TimeUnit.HOURS for window in agg.windows])
                ), (
                    "Detected a snapshot accuracy group by with an hourly aggregation. Resolution with snapshot "
                    "accuracy is not fine enough to allow hourly group bys. Consider removing the `backfill start "
                    "date` param if set or adjusting the aggregation window. "
                    f"input_column: {agg.inputColumn}, windows: {agg.windows}"
                )


_ANY_SOURCE_TYPE = Union[ttypes.Source, ttypes.EventSource, ttypes.EntitySource, ttypes.JoinSource]


def _get_op_suffix(operation, argmap):
    op_str = op_to_str(operation)
    if operation in [
        ttypes.Operation.LAST_K,
        ttypes.Operation.TOP_K,
        ttypes.Operation.FIRST_K,
        ttypes.Operation.BOTTOM_K,
    ]:
        op_name_suffix = op_str[:-2]
        arg_suffix = argmap.get("k")
        return "{}{}".format(op_name_suffix, arg_suffix)
    else:
        return op_str


def get_output_col_names(aggregation):
    base_name = f"{aggregation.inputColumn}_{_get_op_suffix(aggregation.operation, aggregation.argMap)}"
    windowed_names = []
    if aggregation.windows:
        for window in aggregation.windows:
            unit = ttypes.TimeUnit._VALUES_TO_NAMES[window.timeUnit].lower()[0]
            window_suffix = f"{window.length}{unit}"
            windowed_names.append(f"{base_name}_{window_suffix}")
    else:
        windowed_names = [base_name]

    bucketed_names = []
    if aggregation.buckets:
        for bucket in aggregation.buckets:
            bucketed_names.extend([f"{name}_by_{bucket}" for name in windowed_names])
    else:
        bucketed_names = windowed_names

    return bucketed_names


def GroupBy(
    sources: Union[List[_ANY_SOURCE_TYPE], _ANY_SOURCE_TYPE],
    keys: List[str],
    aggregations: Optional[List[ttypes.Aggregation]],
    online: bool = DEFAULT_ONLINE,
    production: bool = DEFAULT_PRODUCTION,
    backfill_start_date: str = None,
    dependencies: List[str] = None,
    env: Dict[str, Dict[str, str]] = None,
    table_properties: Dict[str, str] = None,
    output_namespace: str = None,
    accuracy: ttypes.Accuracy = None,
    lag: int = 0,
    offline_schedule: str = "@daily",
    name: str = None,
    tags: Dict[str, str] = None,
    derivations: List[ttypes.Derivation] = None,
    deprecation_date: str = None,
    **kwargs,
) -> ttypes.GroupBy:
    """

    :param sources:
        can be constructed as entities or events or joinSource::

            import ai.chronon.api.ttypes as chronon
            events = chronon.Source(events=chronon.Events(
                table=YOUR_TABLE,
                topic=YOUR_TOPIC #  <- OPTIONAL for serving
                query=chronon.Query(...)
                isCumulative=False  # <- defaults to false.
            ))
            Or
            entities = chronon.Source(entities=chronon.Entities(
                snapshotTable=YOUR_TABLE,
                mutationTopic=YOUR_TOPIC,
                mutationTable=YOUR_MUTATION_TABLE
                query=chronon.Query(...)
            ))
            or
            joinSource =  chronon.Source(joinSource=chronon.JoinSource(
                join = YOUR_CHRONON_PARENT_JOIN,
                query = chronon.Query(...)
            ))

        Multiple sources can be supplied to backfill the historical values with their respective start and end
        partitions. However, only one source is allowed to be a streaming one.
    :type sources: List[ai.chronon.api.ttypes.Events|ai.chronon.api.ttypes.Entities]
    :param keys:
        List of primary keys that defines the data that needs to be collected in the result table. Similar to the
        GroupBy in the SQL context.
    :type keys: List[String]
    :param aggregations:
        List of aggregations that needs to be computed for the data following the grouping defined by the keys::

            import ai.chronon.api.ttypes as chronon
            aggregations = [
                chronon.Aggregation(input_column="entity", operation=Operation.LAST),
                chronon.Aggregation(input_column="entity", operation=Operation.LAST, windows=[Window(7, TimeUnit.DAYS)])
            ],
    :type aggregations: List[ai.chronon.api.ttypes.Aggregation]
    :param online:
        Should we upload the result data of this conf into the KV store so that we can fetch/serve this GroupBy online.
        Once Online is set to True, you ideally should not change the conf.
    :type online: bool
    :param production:
        This when set can be integrated to trigger alerts. You will have to integrate this flag into your alerting
        system yourself.
    :type production: bool
    :param backfill_start_date:
        Start date from which GroupBy data should be computed. This will determine how far back in time
        that Chronon would go to compute the resultant table and its aggregations.
    :type backfill_start_date: str
    :param dependencies:
        This goes into MetaData.dependencies - which is a list of strings representing the table partitions to wait for.
        Typically used by engines like airflow to create partition sensors.
    :type dependencies: List[str]
    :param env:
        This is a dictionary of "mode name" to dictionary of "env var name" to "env var value"::

            {
                'backfill' : { 'VAR1' : 'VAL1', 'VAR2' : 'VAL2' },
                'upload' : { 'VAR1' : 'VAL1', 'VAR2' : 'VAL2' }
                'streaming' : { 'VAR1' : 'VAL1', 'VAR2' : 'VAL2' }
            }

        These vars then flow into run.py and the underlying spark_submit.sh.
        These vars can be set in other places as well. The priority order (descending) is as below

        1. env vars set while using run.py "VAR=VAL run.py --mode=backfill <name>"
        2. env vars set here in GroupBy's env param
        3. env vars set in `team.json['team.production.<MODE NAME>']`
        4. env vars set in `team.json['default.production.<MODE NAME>']`

    :type env: Dict[str, Dict[str, str]]
    :param table_properties:
        Specifies the properties on output hive tables. Can be specified in teams.json.
    :type table_properties: Dict[str, str]
    :param output_namespace:
        In backfill mode, we will produce data into hive. This represents the hive namespace that the data will be
        written into. You can set this at the teams.json level.
    :type output_namespace: str
    :param accuracy:
        Defines the computing accuracy of the GroupBy.
        If "Snapshot" is selected, the aggregations are computed based on the partition identifier - "ds" time column.
        If "Temporal" is selected, the aggregations are computed based on the event time - "ts" time column.
    :type accuracy: ai.chronon.api.ttypes.SNAPSHOT or ai.chronon.api.ttypes.TEMPORAL
    :param lag:
        Param that goes into customJson. You can pull this out of the json at path "metaData.customJson.lag"
        This is used by airflow integration to pick an older hive partition to wait on.
    :type lag: int
    :param offline_schedule:
        the offline schedule interval for batch jobs. Below is the equivalent of the cron tab commands::

            '@hourly': '0 * * * *',
            '@daily': '0 0 * * *',
            '@weekly': '0 0 * * 0',
            '@monthly': '0 0 1 * *',
            '@yearly': '0 0 1 1 *',

    :type offline_schedule: str
    :param tags:
        Additional metadata that does not directly affect feature computation, but is useful to
        track for management purposes.
    :type tags: Dict[str, str]
    :param derivations:
        Derivation allows arbitrary SQL select clauses to be computed using columns from the output of group by backfill
        output schema. It is supported for offline computations for now.
    :type derivations: List[ai.chronon.api.ttypes.Derivation]
    :param deprecation_date:
        Expected deprecation date of the group by. This is useful to track the deprecation status of the group by.
    :type deprecation_date: str
    :param kwargs:
        Additional properties that would be passed to run.py if specified under additional_args property.
        And provides an option to pass custom values to the processing logic.
    :type kwargs: Dict[str, str]
    :return:
        A GroupBy object containing specified aggregations.
    """
    assert sources, "Sources are not specified"

    agg_inputs = []
    if aggregations is not None:
        agg_inputs = [agg.inputColumn for agg in aggregations]

    required_columns = keys + agg_inputs

    def _sanitize_columns(source: ttypes.Source):
        query = (
            source.entities.query
            if source.entities is not None
            else (source.events.query if source.events is not None else source.joinSource.query)
        )

        if query.selects is None:
            query.selects = {}
        for col in required_columns:
            if col not in query.selects:
                query.selects[col] = col
        if "ts" in query.selects:  # ts cannot be in selects.
            ts = query.selects["ts"]
            del query.selects["ts"]
            if query.timeColumn is None:
                query.timeColumn = ts
            assert query.timeColumn == ts, (
                f"mismatched `ts`: {ts} and `timeColumn`: {query.timeColumn} "
                "in source {source}. Please specify only the `timeColumn`"
            )
        return source

    def _normalize_source(source):
        if isinstance(source, ttypes.EventSource):
            return ttypes.Source(events=source)
        elif isinstance(source, ttypes.EntitySource):
            return ttypes.Source(entities=source)
        elif isinstance(source, ttypes.JoinSource):
            if not source.join.metadata.isSetOutputNamespace():
                source.join.metadata.setOutputNamespace(output_namespace)
            return ttypes.Source(joinSource=source)
        elif isinstance(source, ttypes.Source):
            return source
        else:
            print("unrecognized " + str(source))

    if not isinstance(sources, list):
        sources = [sources]
    sources = [_sanitize_columns(_normalize_source(source)) for source in sources]

    deps = [dep for src in sources for dep in utils.get_dependencies(src, dependencies, lag=lag)]

    kwargs.update({"lag": lag})
    # get caller's filename to assign team
    team = inspect.stack()[1].filename.split("/")[-2]

    column_tags = {}
    if aggregations:
        for agg in aggregations:
            if hasattr(agg, "tags") and agg.tags:
                for output_col in get_output_col_names(agg):
                    column_tags[output_col] = agg.tags
    metadata = {"groupby_tags": tags, "column_tags": column_tags}
    kwargs.update(metadata)

    metadata = ttypes.MetaData(
        name=name,
        online=online,
        production=production,
        outputNamespace=output_namespace,
        customJson=json.dumps(kwargs),
        dependencies=deps,
        modeToEnvMap=env,
        tableProperties=table_properties,
        team=team,
        offlineSchedule=offline_schedule,
        deprecationDate=deprecation_date,
    )

    group_by = ttypes.GroupBy(
        sources=sources,
        keyColumns=keys,
        aggregations=aggregations,
        metaData=metadata,
        backfillStartDate=backfill_start_date,
        accuracy=accuracy,
        derivations=derivations,
    )
    validate_group_by(group_by)
    return group_by



"""
Basic tests for namespace and breaking changes in run.py
"""

#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import argparse
import json
import os
import time

import pytest
from ai.chronon.repo import run

DEFAULT_ENVIRONMENT = os.environ.copy()


@pytest.fixture
def parser():
    """Basic parser for tests relative to the main arguments of run.py"""
    parser = argparse.ArgumentParser()
    args = [
        "repo",
        "conf",
        "mode",
        "env",
        "app-name",
        "chronon-jar",
        "online-jar",
        "online-class",
        "render-info",
        "sub-help",
    ]
    for arg in args:
        parser.add_argument(f"--{arg}")
    run.set_defaults(parser)
    return parser


@pytest.fixture
def test_conf_location():
    """Sample test conf for tests"""
    return "production/joins/sample_team/sample_online_join.v1"


def reset_env(default_env):
    set_keys = os.environ.keys()
    for key in set_keys:
        os.environ.pop(key)
    for k, v in default_env.items():
        os.environ[k] = v


def test_download_jar(monkeypatch, sleepless):
    def mock_cmd(url, path, skip_download):
        return url

    monkeypatch.setattr(time, "sleep", sleepless)
    monkeypatch.setattr(run, "download_only_once", mock_cmd)
    jar_path = run.download_jar(
        "version", jar_type="uber", release_tag=None, spark_version="2.4.0"
    )
    assert jar_path == "/tmp/spark_uber_2.11-version-assembly.jar"
    jar_path = run.download_jar(
        "version", jar_type="uber", release_tag=None, spark_version="3.1.1"
    )
    assert jar_path == "/tmp/spark_uber_2.12-version-assembly.jar"
    with pytest.raises(Exception):
        run.download_jar(
            "version", jar_type="uber", release_tag=None, spark_version="2.1.0"
        )


def test_environment(teams_json, repo, parser, test_conf_location):
    default_environment = DEFAULT_ENVIRONMENT.copy()
    # If nothing is passed.
    run.set_runtime_env(parser.parse_args(args=[]))

    # If repo is passed common_env is loaded.
    reset_env(default_environment)
    run.set_runtime_env(parser.parse_args(args=["--repo", repo]))
    assert os.environ["VERSION"] == "latest"

    # For chronon_metadata_export is passed. APP_NAME should be set.
    reset_env(default_environment)
    run.set_runtime_env(parser.parse_args(args=["--mode", "metadata-export"]))
    assert os.environ["APP_NAME"] == "chronon_metadata_export"

    # If APP_NAME is set, should be respected.
    reset_env(default_environment)
    os.environ["APP_NAME"] = "fake-name"
    run.set_runtime_env(parser.parse_args(args=["--mode", "metadata-export"]))
    assert os.environ["APP_NAME"] == "fake-name"

    # If app_name can be passed from cli.
    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(args=["--mode", "metadata-export", "--app-name", "fake-name"])
    )
    assert os.environ["APP_NAME"] == "fake-name"

    # Check default backfill for a team sets parameters accordingly.
    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(
            args=[
                "--mode",
                "backfill",
                "--conf",
                test_conf_location,
                "--repo",
                repo,
                "--env",
                "production",
                "--online-jar",
                test_conf_location,
            ]
        )
    )
    # from team env.
    assert os.environ["EXECUTOR_CORES"] == "4"
    # from default env.
    assert os.environ["DRIVER_MEMORY"] == "15G"
    # from common env.
    assert os.environ["VERSION"] == "latest"
    # derived from args.
    assert (
        os.environ["APP_NAME"]
        == "chronon_joins_backfill_production_sample_team.sample_online_join.v1"
    )
    # from additional_args
    assert os.environ["CHRONON_CONFIG_ADDITIONAL_ARGS"] == "--step-days 14"

    # Check dev backfill for a team sets parameters accordingly.
    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(
            args=[
                "--mode",
                "backfill",
                "--conf",
                test_conf_location,
                "--repo",
                repo,
                "--online-jar",
                test_conf_location,
            ]
        )
    )
    # from team dev env.
    assert os.environ["EXECUTOR_CORES"] == "2"
    # from team dev env.
    assert os.environ["DRIVER_MEMORY"] == "30G"
    # from default dev env.
    assert os.environ["EXECUTOR_MEMORY"] == "8G"

    # Check conf set environment overrides most.
    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(
            args=[
                "--mode",
                "backfill",
                "--conf",
                "production/joins/sample_team/sample_join.v1",
                "--repo",
                repo,
                "--env",
                "production",
            ]
        )
    )
    # from conf env.
    assert os.environ["EXECUTOR_MEMORY"] == "9G"

    # Bad conf location raises error.
    with pytest.raises(Exception):
        reset_env(default_environment)
        run.set_runtime_env(
            parser.parse_args(
                args=[
                    "--mode",
                    "backfill",
                    "--conf",
                    "joins/sample_team/sample_join.v1",
                    "--repo",
                    repo,
                ]
            )
        )

    # Check metadata export run.py
    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(
            args=[
                "--mode",
                "metadata-export",
                "--conf",
                "production/joins//",
                "--repo",
                repo,
            ]
        )
    )
    # without conf still works.
    assert os.environ["APP_NAME"] == "chronon_joins_metadata_export"

    reset_env(default_environment)
    run.set_runtime_env(
        parser.parse_args(
            args=[
                "--mode",
                "metadata-upload",
                "--conf",
                "production/joins//",
                "--repo",
                repo,
            ]
        )
    )
    assert os.environ["APP_NAME"] == "chronon_joins_metadata_upload"
    reset_env(default_environment)


def test_property_default_update(repo, parser, test_conf_location):
    reset_env(DEFAULT_ENVIRONMENT.copy())
    assert "VERSION" not in os.environ
    args, _ = parser.parse_known_args(
        args=["--mode", "backfill", "--conf", test_conf_location, "--repo", repo]
    )
    assert args.version is None
    run.set_runtime_env(args)
    assert "VERSION" in os.environ
    assert args.version is None
    run.set_defaults(parser)
    reparsed, _ = parser.parse_known_args(
        args=["--mode", "backfill", "--conf", test_conf_location, "--repo", repo]
    )
    assert reparsed.version is not None


def test_render_info_setting_update(repo, parser, test_conf_location):
    default_environment = DEFAULT_ENVIRONMENT.copy()

    run.set_defaults(parser)
    args, _ = parser.parse_known_args(
        args=["--mode", "info", "--conf", test_conf_location, "--repo", repo]
    )
    run.set_defaults(parser)
    assert args.render_info == os.path.join(".", run.RENDER_INFO_DEFAULT_SCRIPT)

    reset_env(default_environment)
    run.set_runtime_env(args)
    os.environ["CHRONON_REPO_PATH"] = repo
    run.set_defaults(parser)
    args, _ = parser.parse_known_args(
        args=["--mode", "info", "--conf", test_conf_location, "--repo", repo]
    )
    assert args.render_info == os.path.join(repo, run.RENDER_INFO_DEFAULT_SCRIPT)

    reset_env(default_environment)
    run.set_defaults(parser)
    somewhere = "/tmp/somewhere/script.py"
    args, _ = parser.parse_known_args(
        args=[
            "--mode",
            "info",
            "--conf",
            test_conf_location,
            "--render-info",
            somewhere,
        ]
    )
    assert args.render_info == somewhere


def test_render_info(repo, parser, test_conf_location, monkeypatch):
    actual_cmd = None

    def mock_check_call(cmd):
        nonlocal actual_cmd
        actual_cmd = cmd
        return cmd

    def mock_exists(_):
        return True

    monkeypatch.setattr(run, "check_call", mock_check_call)
    monkeypatch.setattr(os.path, "exists", mock_exists)
    run.set_defaults(parser)
    args, _ = parser.parse_known_args(
        args=["--mode", "info", "--conf", test_conf_location, "--repo", repo]
    )

    args.args = _
    runner = run.Runner(args, "some.jar")
    runner.run()

    assert run.RENDER_INFO_DEFAULT_SCRIPT in actual_cmd


def test_streaming_client(repo, parser, test_online_group_by, monkeypatch):
    """Test mode compiles properly and uses the same app name by default, killing if necessary."""
    calls = []

    def mock_check_call(cmd):
        nonlocal calls
        calls += [cmd]
        return cmd

    def mock_check_output(cmd):
        print(cmd)
        return "[]".encode("utf8")

    monkeypatch.setattr(run, "check_output", mock_check_output)
    monkeypatch.setattr(run, "check_call", mock_check_call)
    run.set_defaults(parser)
    # Follow the same flow as __main__: Do a first pass (no env), do a second pass and run.
    pre_parse_args, _ = parser.parse_known_args(
        args=["--mode", "streaming", "--conf", test_online_group_by, "--repo", repo]
    )
    run.set_runtime_env(pre_parse_args)
    run.set_defaults(parser)
    parse_args, _ = parser.parse_known_args(
        args=["--mode", "streaming", "--conf", test_online_group_by, "--repo", repo]
    )
    parse_args.args = ""
    runner = run.Runner(parse_args, "some.jar")
    runner.run()
    streaming_app_name = runner.app_name
    # Repeat for streaming-client
    pre_parse_args, _ = parser.parse_known_args(
        args=[
            "--mode",
            "streaming-client",
            "--conf",
            test_online_group_by,
            "--repo",
            repo,
        ]
    )
    run.set_runtime_env(pre_parse_args)
    run.set_defaults(parser)
    parse_args, _ = parser.parse_known_args(
        args=[
            "--mode",
            "streaming-client",
            "--conf",
            test_online_group_by,
            "--repo",
            repo,
        ]
    )
    parse_args.args = ""
    runner = run.Runner(parse_args, "some.jar")
    runner.run()
    assert streaming_app_name == runner.app_name

    # Check job its not killed if found and submitted by a different user.
    def mock_check_output_with_app_other_user(cmd):
        return json.dumps(
            {
                "app_name": streaming_app_name,
                "kill_cmd": "<kill app cmd>",
                "user": "notcurrent",
            }
        ).encode("utf8")

    monkeypatch.setattr(run, "check_output", mock_check_output_with_app_other_user)
    assert "<kill app cmd>" not in calls
    runner = run.Runner(parse_args, "some.jar")
    with pytest.raises(RuntimeError):
        runner.run()


def test_split_date_range():
    start_date = "2022-01-01"
    end_date = "2022-01-11"
    parallelism = 5
    expected_result = [
        ("2022-01-01", "2022-01-02"),
        ("2022-01-03", "2022-01-04"),
        ("2022-01-05", "2022-01-06"),
        ("2022-01-07", "2022-01-08"),
        ("2022-01-09", "2022-01-11"),
    ]

    result = run.split_date_range(start_date, end_date, parallelism)
    assert result == expected_result



#!/usr/bin/env python
# tool to materialize feature_sources and feature_sets into thrift configurations
# that chronon jobs can consume


#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import json
import logging
import os
import textwrap
from typing import Optional, Tuple, Type, Union

import ai.chronon.api.ttypes as api
import ai.chronon.repo.extract_objects as eo
import ai.chronon.utils as utils
import click
from ai.chronon.api.ttypes import GroupBy, Join, StagingQuery
from ai.chronon.repo import (
    GROUP_BY_FOLDER_NAME,
    JOIN_FOLDER_NAME,
    STAGING_QUERY_FOLDER_NAME,
    TEAMS_FILE_PATH,
    dependency_tracker,
    teams,
)
from ai.chronon.repo.serializer import thrift_simple_json_protected
from ai.chronon.repo.validator import (
    ChrononRepoValidator,
    get_group_by_output_columns,
    get_join_output_columns,
)

# This is set in the main function -
# from command line or from env variable during invocation
FOLDER_NAME_TO_CLASS = {
    GROUP_BY_FOLDER_NAME: GroupBy,
    JOIN_FOLDER_NAME: Join,
    STAGING_QUERY_FOLDER_NAME: StagingQuery,
}

DEFAULT_TEAM_NAME = "default"


def get_folder_name_from_class_name(class_name):
    return {v.__name__: k for k, v in FOLDER_NAME_TO_CLASS.items()}[class_name]


@click.command()
@click.option("--chronon_root", envvar="CHRONON_ROOT", help="Path to the root chronon folder", default=os.getcwd())
@click.option(
    "--input_path",
    "--conf",
    "input_path",
    help="Relative Path to the root chronon folder, which contains the objects to be serialized",
    required=True,
)
@click.option(
    "--output_root",
    help="Relative Path to the root chronon folder, to where the serialized output should be written",
    default="production",
)
@click.option("--debug", help="debug mode", is_flag=True)
@click.option("--force-overwrite", help="Force overwriting existing materialized conf.", is_flag=True)
@click.option("--feature-display", help="Print out the features list created by the conf.", is_flag=True)
@click.option(
    "--table-display",
    help="Print out the list of tables that are materialized per job modes in this conf.",
    is_flag=True,
)
def extract_and_convert(chronon_root, input_path, output_root, debug, force_overwrite, feature_display, table_display):
    """
    CLI tool to convert Python chronon GroupBy's, Joins and Staging queries into their thrift representation.
    The materialized objects are what will be submitted to spark jobs - driven by airflow, or by manual user testing.
    """
    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    _print_highlighted("Using chronon root path", chronon_root)
    chronon_root_path = os.path.expanduser(chronon_root)
    utils.chronon_root_path = chronon_root_path
    path_split = input_path.split("/")
    obj_folder_name = path_split[0]
    obj_class = FOLDER_NAME_TO_CLASS[obj_folder_name]
    full_input_path = os.path.join(chronon_root_path, input_path)
    _print_highlighted(f"Input {obj_folder_name} from", full_input_path)
    assert os.path.exists(full_input_path), f"Input Path: {full_input_path} doesn't exist"
    if os.path.isdir(full_input_path):
        results = eo.from_folder(chronon_root_path, full_input_path, obj_class, log_level=log_level)
    elif os.path.isfile(full_input_path):
        assert full_input_path.endswith(".py"), f"Input Path: {input_path} isn't a python file"
        results = eo.from_file(chronon_root_path, full_input_path, obj_class, log_level=log_level)
    else:
        raise Exception(f"Input Path: {full_input_path}, isn't a file or a folder")
    validator = ChrononRepoValidator(chronon_root_path, output_root, log_level=log_level)
    extra_online_or_gb_backfill_enabled_group_bys = {}
    extra_dependent_group_bys_to_materialize = {}
    extra_dependent_joins_to_materialize = {}
    num_written_objs = 0
    full_output_root = os.path.join(chronon_root_path, output_root)
    teams_path = os.path.join(chronon_root_path, TEAMS_FILE_PATH)
    entity_dependency_tracker = dependency_tracker.ChrononEntityDependencyTracker(
        chronon_root_path=full_output_root,
        log_level=log_level,
    )
    for name, obj in results.items():
        team_name = name.split(".")[0]
        _set_team_level_metadata(obj, teams_path, team_name)
        _set_templated_values(obj, obj_class, teams_path, team_name)
        if _write_obj(full_output_root, validator, name, obj, log_level, force_overwrite, force_overwrite):
            num_written_objs += 1

            if feature_display:
                _print_features(obj, obj_class)

            if table_display:
                _print_tables(obj, obj_class)

            # In case of join, we need to materialize the following underlying group_bys
            # 1. group_bys whose online param is set
            # 2. group_bys whose backfill_start_date param is set.
            if obj_class is Join:
                online_group_bys = {}
                offline_backfill_enabled_group_bys = {}
                offline_gbs = []  # gather list to report errors
                for jp in obj.joinParts:
                    if jp.groupBy.metaData.online:
                        online_group_bys[jp.groupBy.metaData.name] = jp.groupBy
                    elif jp.groupBy.backfillStartDate:
                        offline_backfill_enabled_group_bys[jp.groupBy.metaData.name] = jp.groupBy
                    else:
                        offline_gbs.append(jp.groupBy.metaData.name)
                logging.debug(
                    f"Online GroupBys: {online_group_bys.keys()}, "
                    f"Offline GroupBys with backfill enabled: {offline_backfill_enabled_group_bys.keys()}, "
                    f"Offline GroupBys: {offline_gbs}"
                )
                extra_online_or_gb_backfill_enabled_group_bys.update(
                    {**online_group_bys, **offline_backfill_enabled_group_bys}
                )
                if obj.metaData.online:
                    group_bys_not_online = offline_gbs + list(offline_backfill_enabled_group_bys.keys())
                    assert not group_bys_not_online, (
                        "You must make all dependent GroupBys `online` if you want to make your join [{}] `online`."
                        " You can do this by passing the `online=True` argument to the GroupBy constructor."
                        " Fix the following: {}".format(obj.metaData.name, group_bys_not_online)
                    )

            new_group_bys, new_joins = _handle_dependent_configurations(
                chronon_root_path, entity_dependency_tracker, full_output_root, name, obj, obj_class, log_level
            )

            # Update the accumulated dictionaries with the new entries
            extra_dependent_group_bys_to_materialize.update(new_group_bys)
            extra_dependent_joins_to_materialize.update(new_joins)

    if not force_overwrite:
        dependencies = {}
        dependencies.update({**extra_dependent_group_bys_to_materialize, **extra_dependent_joins_to_materialize})
        assert not dependencies, (
            "You must also materialize all dependent GroupBys or Joins."
            " You can do this by passing the --force-overwrite flag to the compile.py command."
            " Detected dependencies are as follows: {}".format(sorted(dependencies.keys()))
        )

    if extra_online_or_gb_backfill_enabled_group_bys or extra_dependent_group_bys_to_materialize:
        extra_online_or_gb_backfill_enabled_group_bys.update(extra_dependent_group_bys_to_materialize)
        _handle_extra_conf_objects_to_materialize(
            conf_objs=extra_online_or_gb_backfill_enabled_group_bys,
            force_overwrite=force_overwrite,
            full_output_root=full_output_root,
            teams_path=teams_path,
            validator=validator,
            log_level=log_level,
        )
    if extra_dependent_joins_to_materialize:
        _handle_extra_conf_objects_to_materialize(
            conf_objs=extra_dependent_joins_to_materialize,
            force_overwrite=force_overwrite,
            full_output_root=full_output_root,
            teams_path=teams_path,
            validator=validator,
            log_level=log_level,
            is_gb=False,
        )
    if num_written_objs > 0:
        print(f"Successfully wrote {num_written_objs} {(obj_class).__name__} objects to {full_output_root}")


def _handle_dependent_configurations(
    chronon_root_path: str,
    entity_dependency_tracker: dependency_tracker.ChrononEntityDependencyTracker,
    full_output_root: str,
    name: str,
    obj: object,
    obj_class: Type[Union[Join, GroupBy]],
    log_level=logging.INFO,
) -> Tuple[dict, dict]:
    # Initialize local dictionaries to store new entries to be added
    new_group_bys_to_materialize = {}
    new_joins_to_materialize = {}

    output_file_path = _get_relative_materialized_file_path(full_output_root, name, obj)
    downstreams = entity_dependency_tracker.check_downstream(output_file_path)

    for downstream in downstreams:
        if obj_class is Join:
            downstream_class = GroupBy
        elif obj_class is GroupBy:
            downstream_class = Join
        else:
            continue
        conf_result = _get_conf_file_path(downstream, downstream_class)
        if conf_result is None:
            continue

        conf_var, conf_file_path = conf_result
        conf_obj = eo.from_file(
            chronon_root_path,
            os.path.join(chronon_root_path, conf_file_path),
            downstream_class,
            log_level=log_level,
        )

        obj_to_materialize = conf_obj[downstream]
        if downstream_class is GroupBy:
            new_group_bys_to_materialize[obj_to_materialize.metaData.name] = obj_to_materialize
        elif downstream_class is Join:
            new_joins_to_materialize[obj_to_materialize.metaData.name] = obj_to_materialize

    # Return new entries to be added to the main dictionaries
    return new_group_bys_to_materialize, new_joins_to_materialize


def _get_relative_materialized_file_path(full_output_root: str, name: str, obj: object) -> str:
    _, _, abs_output_file_path = _construct_output_file_name(full_output_root, name, obj)
    output_file_path = abs_output_file_path.replace(full_output_root + os.sep, "")
    return output_file_path


def _get_conf_file_path(downstream: str, downstream_class: Type[Union[Join, GroupBy]]) -> Optional[Tuple[str, str]]:
    parts = downstream.split(".")

    # This may happen for scenarios where group_by is defined within join file itself.
    # So no separate group_by file is present. In such cases, compiling the join will handle the materialize
    # for group_by.
    if len(parts) < 3:
        return None

    team_name = parts[0]
    conf_var = parts[-1]
    conf_name = parts[-2] + ".py"

    # Nested path between team_name and conf_name
    nested_path = parts[1:-2] if len(parts) > 3 else []

    # Construct the path for configuration file
    conf_file_path = os.path.join(team_name, *nested_path, conf_name)

    # Return the appropriate path based on downstream_class
    if downstream_class is Join:
        return conf_var, os.path.join(JOIN_FOLDER_NAME, conf_file_path)
    elif downstream_class is GroupBy:
        return conf_var, os.path.join(GROUP_BY_FOLDER_NAME, conf_file_path)
    else:
        return None  # Handle other cases or errors if needed


def _print_features(obj: Union[Join, GroupBy], obj_class: Type[Union[Join, GroupBy]]) -> None:
    if obj_class is Join:
        _print_features_names("Output Join Features", get_join_output_columns(obj))
    if obj_class is GroupBy:
        _print_features_names("Output GroupBy Features", get_group_by_output_columns(obj))


def _print_tables(obj: Union[Join, GroupBy], obj_class: Type[Union[Join, GroupBy]]) -> None:
    tables = utils.get_modes_tables(obj)
    if obj_class is Join:
        _print_modes_tables("Output Join Tables", tables)
    if obj_class is GroupBy:
        _print_modes_tables("Output GroupBy Tables", tables)


def _handle_extra_conf_objects_to_materialize(
    conf_objs,
    force_overwrite: bool,
    full_output_root: str,
    teams_path: str,
    validator: ChrononRepoValidator,
    log_level=logging.INFO,
    is_gb=True,
) -> None:
    num_written_objs = 0
    # load materialized joins to validate the additional conf objects against.
    validator.load_objs()
    for name, obj in conf_objs.items():
        team_name = name.split(".")[0]
        _set_team_level_metadata(obj, teams_path, team_name)
        if _write_obj(
            full_output_root,
            validator,
            name,
            obj,
            log_level,
            force_compile=True,
            force_overwrite=force_overwrite,
        ):
            num_written_objs += 1
    print(f"Successfully wrote {num_written_objs} {'GroupBy' if is_gb else 'Join'} objects to {full_output_root}")


def _set_team_level_metadata(obj: object, teams_path: str, team_name: str):
    namespace = teams.get_team_conf(teams_path, team_name, "namespace")
    table_properties = teams.get_team_conf(teams_path, team_name, "table_properties")
    obj.metaData.outputNamespace = obj.metaData.outputNamespace or namespace
    obj.metaData.tableProperties = obj.metaData.tableProperties or table_properties
    obj.metaData.team = team_name

    # set metadata for JoinSource
    if isinstance(obj, api.GroupBy):
        for source in obj.sources:
            if source.joinSource:
                _set_team_level_metadata(source.joinSource.join, teams_path, team_name)


def __fill_template(table, obj, namespace):
    if table:
        table = table.replace("{{ logged_table }}", utils.log_table_name(obj, full_name=True))
        table = table.replace("{{ db }}", namespace)
    return table


def _set_templated_values(obj, cls, teams_path, team_name):
    namespace = teams.get_team_conf(teams_path, team_name, "namespace")
    if cls == api.Join and obj.bootstrapParts:
        for bootstrap in obj.bootstrapParts:
            bootstrap.table = __fill_template(bootstrap.table, obj, namespace)
        if obj.metaData.dependencies:
            obj.metaData.dependencies = [__fill_template(dep, obj, namespace) for dep in obj.metaData.dependencies]
    if cls == api.Join and obj.labelPart:
        obj.labelPart.metaData.dependencies = [
            label_dep.replace("{{ join_backfill_table }}", utils.output_table_name(obj, full_name=True))
            for label_dep in obj.labelPart.metaData.dependencies
        ]


def _write_obj(
    full_output_root: str,
    validator: ChrononRepoValidator,
    name: str,
    obj: object,
    log_level: int,
    force_compile: bool = False,
    force_overwrite: bool = False,
) -> bool:
    """
    Returns True if the object is successfully written.
    """
    file_name, obj_class, output_file = _construct_output_file_name(full_output_root, name, obj)
    class_name = obj_class.__name__
    team_name = name.split(".")[0]
    _print_highlighted(f"{class_name} Team", team_name)
    _print_highlighted(f"{class_name} Name", file_name)
    skip_reasons = validator.can_skip_materialize(obj)
    if not force_compile and skip_reasons:
        reasons = ", ".join(skip_reasons)
        _print_warning(f"Skipping {class_name} {file_name}: {reasons}")
        if os.path.exists(output_file):
            _print_warning(f"old file exists for skipped config: {output_file}")
        return False
    validation_errors = validator.validate_obj(obj)
    if validation_errors:
        _print_error(f"Could not write {class_name} {file_name}", ", ".join(validation_errors))
        return False
    if force_overwrite:
        _print_warning(f"Force overwrite {class_name} {file_name}")
    elif not validator.safe_to_overwrite(obj):
        _print_warning(f"Cannot overwrite {class_name} {file_name} with existing online conf")
        return False
    _write_obj_as_json(file_name, obj, output_file, obj_class)
    return True


def _construct_output_file_name(full_output_root: str, name: str, obj: object) -> Tuple[str, Type, str]:
    team_name = name.split(".")[0]
    obj_class = type(obj)
    class_name = obj_class.__name__
    file_name = name.split(".", 1)[1]
    obj_folder_name = get_folder_name_from_class_name(class_name)
    output_path = os.path.join(full_output_root, obj_folder_name, team_name)
    output_file = os.path.join(output_path, file_name)
    return file_name, obj_class, output_file


def _write_obj_as_json(name: str, obj: object, output_file: str, obj_class: type):
    class_name = obj_class.__name__
    output_folder = os.path.dirname(output_file)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    assert os.path.isdir(output_folder), f"{output_folder} isn't a folder."
    assert hasattr(obj, "name") or hasattr(
        obj, "metaData"
    ), f"Can't serialize objects without the name attribute for object {name}"
    with open(output_file, "w") as f:
        _print_highlighted(f"Writing {class_name} to", output_file)
        f.write(thrift_simple_json_protected(obj, obj_class))


def _print_highlighted(left, right):
    # print in blue.
    print(f"{left:>25} - \u001b[34m{right}\u001b[0m")


def _print_features_names(left, right):
    # Print in green and separate lines.
    print(f"{left:>25} - \u001b[32m{right}\u001b[0m")


def _print_modes_tables(left, right):
    text = textwrap.indent(json.dumps(right, indent=2), " " * 27)
    print(f"{left:>25} - \u001b[32m\n{text}\u001b[0m")


def _print_error(left, right):
    # print in red.
    print(f"\033[91m{left:>25} - \033[1m{right}\033[00m")


def _print_warning(string):
    # print in yellow - \u001b[33m
    print(f"\u001b[33m{string}\u001b[0m")


if __name__ == "__main__":
    extract_and_convert()




#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from ai.chronon.join import Join
from ai.chronon.group_by import GroupBy
from ai.chronon.api import ttypes as api

import pytest
import json


def event_source(table):
    """
    Sample left join
    """
    return api.Source(
        events=api.EventSource(
            table=table,
            query=api.Query(
                startPartition="2020-04-09",
                selects={
                    "subject": "subject_sql",
                    "event_id": "event_sql",
                },
                timeColumn="CAST(ts AS DOUBLE)",
            ),
        ),
    )


def right_part(source):
    """
    Sample Agg
    """
    return api.JoinPart(
        groupBy=api.GroupBy(
            sources=[source],
            keyColumns=["subject"],
            aggregations=[],
            accuracy=api.Accuracy.SNAPSHOT,
            backfillStartDate="2020-04-09",
        ),
    )


def test_deduped_dependencies():
    """
    Check left and right dependencies are deduped in metadata.
    """
    join = Join(
        left=event_source("sample_namespace.sample_table"),
        right_parts=[right_part(event_source("sample_namespace.another_table"))])
    assert len(join.metaData.dependencies) == 2

    join = Join(
        left=event_source("sample_namespace.sample_table"),
        right_parts=[right_part(event_source("sample_namespace.sample_table"))])
    assert len(join.metaData.dependencies) == 1


def test_additional_args_to_custom_json():
    join = Join(
        left=event_source("sample_namespace.sample_table"),
        right_parts=[right_part(event_source("sample_namespace.sample_table"))],
        team_override="some_other_team_value"
    )
    assert json.loads(join.metaData.customJson)['team_override'] == "some_other_team_value"


def test_dependencies_propagation():
    gb1 = GroupBy(
        sources=[event_source("table_1")],
        keys=["subject"],
        aggregations=[],
    )
    gb2 = GroupBy(
        sources=[event_source("table_2")],
        keys=["subject"],
        aggregations=[],
        dependencies=["table_2/ds={{ ds }}/key=value"]
    )
    join = Join(
        left=event_source("left_1"),
        right_parts=[api.JoinPart(gb1), api.JoinPart(gb2)]
    )

    actual = [
        (json.loads(dep)["name"], json.loads(dep)["spec"])
        for dep in join.metaData.dependencies
    ]
    expected = [
        ("wait_for_left_1_ds", "left_1/ds={{ ds }}"),
        ("wait_for_table_1_ds", "table_1/ds={{ ds }}"),
        ("wait_for_table_2_ds_ds_key_value", "table_2/ds={{ ds }}/key=value")
    ]
    assert expected == actual



"""Object for checking the downstream dependency of an entity in the Chronon repository.
"""

#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
import logging
import os
from ai.chronon.logger import get_logger
from typing import Set
from ai.chronon.api.ttypes import \
    GroupBy, Join
from ai.chronon.repo import JOIN_FOLDER_NAME, \
    GROUP_BY_FOLDER_NAME
from ai.chronon.repo.validator import extract_json_confs


class ChrononEntityDependencyTracker(object):
    def __init__(self, chronon_root_path: str, log_level=logging.INFO):
        self.logger = get_logger(log_level)
        self.chronon_root_path = chronon_root_path

    def extract_conf(
        self, obj_class: type, path: str
    ) -> object:
        obj = extract_json_confs(
            obj_class,
            os.path.join(self.chronon_root_path, path))
        if len(obj) == 0:
            raise Exception(f"No {obj_class} found in {path}")
        elif len(obj) > 1:
            raise Exception(f"Multiple {obj_class} found in {path}")
        return obj[0]

    def check_join_downstream(self, conf_path) -> Set[str]:
        join = self.extract_conf(Join, conf_path)
        join_name = join.metaData.name
        downstreams = set()
        group_bys = extract_json_confs(
            GroupBy,
            os.path.join(self.chronon_root_path, GROUP_BY_FOLDER_NAME))
        for group_by in group_bys:
            for source in group_by.sources:
                if source.joinSource and source.joinSource.join.metaData.name == join_name:
                    downstreams.add(group_by.metaData.name)
        return downstreams

    def check_group_by_downstream(self, conf_path) -> Set[str]:
        group_by = self.extract_conf(GroupBy, conf_path)
        group_by_name = group_by.metaData.name
        downstreams = set()
        joins = extract_json_confs(
            Join,
            os.path.join(self.chronon_root_path, JOIN_FOLDER_NAME))
        for join in joins:
            for join_part in join.joinParts:
                if join_part.groupBy.metaData.name == group_by_name:
                    downstreams.add(join.metaData.name)
        return downstreams

    def check_downstream(self, conf_path: str) -> Set[str]:
        if "joins" in conf_path:
            downstream = self.check_join_downstream(conf_path)
        elif "group_bys" in conf_path:
            downstream = self.check_group_by_downstream(conf_path)
        else:
            raise Exception(f"Invalid conf path: {conf_path}")
        return downstream



"""
Test dependency tracker
"""

#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import pytest

from ai.chronon.repo import dependency_tracker


@pytest.fixture
def test_dependency_tracker():
    return dependency_tracker.ChrononEntityDependencyTracker(
        chronon_root_path='test/sample/production'
    )


def test_group_by_dependency(test_dependency_tracker):
    conf_path = 'group_bys/sample_team/event_sample_group_by.v1'
    downstream = test_dependency_tracker.check_downstream(conf_path)
    assert (len(downstream) == 8)
    assert ('sample_team.sample_join_derivation.v1' in downstream)


def test_join_dependency(test_dependency_tracker):
    conf_path = 'joins/sample_team/sample_chaining_join.parent_join'
    downstream = test_dependency_tracker.check_downstream(conf_path)
    assert (len(downstream) == 1)
    assert ('sample_team.sample_chaining_group_by' in downstream)



"""
Run the flow for materialize.
"""

#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import os

import pytest
from ai.chronon.repo.compile import extract_and_convert
from click.testing import CliRunner

CURRENT_FILE_DIR = os.path.dirname(os.path.abspath(__file__))


# Utility functions
def _get_full_file_path(relative_path):
    """Utility function to get the full file path based on a relative path."""
    return os.path.join(CURRENT_FILE_DIR, relative_path)


def _invoke_cli_with_params(runner, input_path):
    """Invoke the CLI command with consistent options and specified input_path."""
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            f"--input_path={input_path}",
            "--debug",
        ],
    )
    return result


def _assert_file_exists(full_file_path, message):
    """Assert that a file exists at the specified path."""
    assert os.path.isfile(full_file_path), message


@pytest.fixture
def specific_setup():
    # This setup code will only run for tests that request this fixture
    yield  # This is where the testing happens
    # Teardown phase: Cleanup after tests
    current_file_dir = os.path.dirname(os.path.abspath(__file__))
    files_to_clean = [
        "sample/production/group_bys/unit_test/event_sample_group_by.v1",
        "sample/production/group_bys/unit_test/entity_sample_group_by.require_backfill",
        "sample/production/joins/unit_test/sample_online_join.v1",
        "sample/production/joins/unit_test/sample_join.v1",
        "sample/production/joins/unit_test/sample_online_join_with_gb_not_online.v1",
    ]

    for relative_path in files_to_clean:
        full_path = os.path.join(current_file_dir, relative_path)
        if os.path.exists(full_path):
            os.remove(full_path)


def test_basic_compile():
    runner = CliRunner()
    result = runner.invoke(extract_and_convert, ["--chronon_root=test/sample", "--input_path=joins/sample_team/"])
    assert result.exit_code == 0
    result = runner.invoke(extract_and_convert, ["--chronon_root=test/sample", "--input_path=joins/sample_team"])
    assert result.exit_code == 0
    result = runner.invoke(
        extract_and_convert, ["--chronon_root=test/sample", "--input_path=joins/sample_team/sample_join.py"]
    )
    assert result.exit_code == 0
    result = runner.invoke(
        extract_and_convert,
        ["--chronon_root=test/sample", "--input_path=group_bys/sample_team/sample_deprecation_group_by.py"],
    )
    assert result.exit_code == 0


def test_debug_compile():
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert, ["--chronon_root=test/sample", "--input_path=joins/sample_team/", "--debug"]
    )
    assert result.exit_code == 0


def test_failed_compile():
    """
    Should fail as it fails to find teams.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--input_path=joins/sample_team/",
        ],
    )
    assert result.exit_code != 0


def test_failed_compile_online_join_not_online_gb(specific_setup):
    """
    Raises an error when compiling an online join that references a group_by that is not online.
    """
    runner = CliRunner()
    result = _invoke_cli_with_params(runner, "joins/unit_test/sample_online_join_with_gb_not_online.py")

    error_message_expected = "Fix the following: ['unit_test.entity_sample_group_by.v1']"
    assert result.exit_code != 0, "Command unexpectedly succeeded, but an error was expected."
    assert error_message_expected in str(
        result.exception
    ), f"Got a different message than expected.: {result.exception}"
    assert isinstance(
        result.exception, AssertionError
    ), "Expected an AssertionError, but got a different exception or no exception."


def test_compiling_online_join_compiles_all_online_gb(specific_setup):
    """
    Test that compiling an online join correctly materializes all online group_bys.
    """
    runner = CliRunner()
    result = _invoke_cli_with_params(runner, "joins/unit_test/sample_online_join.py")
    assert result.exit_code == 0

    full_file_path = _get_full_file_path("sample/production/group_bys/unit_test/event_sample_group_by.v1")
    _assert_file_exists(full_file_path, "Expected the group_by to be materialized, but it was not.")


def test_compiling_join_compiles_online_bsd_gb(specific_setup):
    """
    Test that compiling a join correctly materializes online group_bys and those with backfillStartDate set.
    """
    runner = CliRunner()
    result = _invoke_cli_with_params(runner, "joins/unit_test/sample_join.py")
    assert result.exit_code == 0

    paths_to_check = [
        "sample/production/group_bys/unit_test/event_sample_group_by.v1",
        "sample/production/group_bys/unit_test/entity_sample_group_by.require_backfill",
    ]
    for path in paths_to_check:
        full_file_path = _get_full_file_path(path)
        _assert_file_exists(full_file_path, f"Expected {os.path.basename(path)} to be materialized, but it was not.")


def test_failed_compile_missing_input_column():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            "--input_path=group_bys/sample_team/sample_group_by_missing_input_column.py",
            "--debug",
        ],
    )
    assert result.exit_code != 0


def test_failed_compile_when_dependent_join_detected():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        ["--chronon_root=test/sample", "--input_path=group_bys/sample_team/event_sample_group_by.py"],
    )
    assert result.exit_code != 0
    error_message_expected = "Detected dependencies are as follows: ['sample_team.sample_chaining_join.parent_join', 'sample_team.sample_join_bootstrap.v1', 'sample_team.sample_join_bootstrap.v2', 'sample_team.sample_join_derivation.v1', 'sample_team.sample_join_with_derivations_on_external_parts.v1', 'sample_team.sample_label_join.v1', 'sample_team.sample_label_join_with_agg.v1', 'sample_team.sample_online_join.v1']"
    actual_exception_message = str(result.exception).strip().lower()
    error_message_expected = error_message_expected.strip().lower()
    assert (
        error_message_expected in actual_exception_message
    ), f"Got a different message than expected {actual_exception_message}"
    assert isinstance(
        result.exception, AssertionError
    ), "Expected an AssertionError, but got a different exception or no exception."


def test_detected_dependent_joins_materialized():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            "--input_path=group_bys/sample_team/event_sample_group_by.py",
            "--force-overwrite",
        ],
    )
    assert result.exit_code == 0
    expected_message = "Successfully wrote 8 Join objects to test/sample/production".strip().lower()
    actual_message = str(result.output).strip().lower()
    assert expected_message in actual_message, f"Got a different message than expected {actual_message}"


def test_failed_compile_when_dependent_groupby_detected():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            "--input_path=joins/unit_test/sample_parent_join.py",
        ],
    )
    assert result.exit_code != 0
    error_message_expected = (
        "Detected dependencies are as follows: ['unit_test.sample_chaining_group_by.chaining_group_by_v1']"
    )
    actual_exception_message = str(result.exception).strip().lower()
    error_message_expected = error_message_expected.strip().lower()
    assert (
        error_message_expected in actual_exception_message
    ), f"Got a different message than expected {actual_exception_message}"
    assert isinstance(
        result.exception, AssertionError
    ), "Expected an AssertionError, but got a different exception or no exception."


def test_detected_dependent_group_bys_materialized():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            "--input_path=joins/unit_test/sample_parent_join.py",
            "--force-overwrite",
        ],
    )
    assert result.exit_code == 0
    expected_message = "Successfully wrote 2 GroupBy objects to test/sample/production".strip().lower()
    actual_message = str(result.output).strip().lower()
    assert expected_message in actual_message, f"Got a different message than expected {actual_message}"


def test_detected_dependent_nested_joins():
    """
    Should raise errors as we are trying to create aggregations without input column.
    """
    runner = CliRunner()
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            "--input_path=group_bys/unit_test/user/sample_nested_group_by.py",
            "--force-overwrite",
        ],
    )
    assert result.exit_code == 0
    expected_message = "Successfully wrote 1 Join objects to test/sample/production".strip().lower()
    actual_message = str(result.output).strip().lower()
    assert expected_message in actual_message, f"Got a different message than expected {actual_message}"


def test_compile_table_display():
    """
    Test that compiling an online join correctly materializes all online group_bys.
    """
    runner = CliRunner()
    input_path = f"joins/sample_team/sample_join_with_derivations_on_external_parts.py"
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            f"--input_path={input_path}",
            "--table-display",
        ],
    )

    assert "Output Join Tables" in result.output
    assert result.exit_code == 0


def test_compile_feature_display():
    """
    Test that compiling an online join correctly materializes all online group_bys.
    """
    runner = CliRunner()
    input_path = f"joins/sample_team/sample_join_with_derivations_on_external_parts.py"
    result = runner.invoke(
        extract_and_convert,
        [
            "--chronon_root=test/sample",
            f"--input_path={input_path}",
            "--feature-display",
        ],
    )

    assert "Output Join Features" in result.output
    assert result.exit_code == 0
